# CLAUDE.md

This file provides guidance to Claude Code (claude.ai/code) when working with code in this repository.

## Project Overview

This is an enhanced CVE-SIGMA Auto Generator that automatically processes comprehensive CVE data and generates SIGMA rules for threat detection. The application now supports:

1. **Bulk NVD Data Processing**: Downloads and processes complete NVD JSON datasets (2002-2025)
2. **nomi-sec PoC Integration**: Uses curated PoC data from github.com/nomi-sec/PoC-in-GitHub
3. **Enhanced SIGMA Rule Generation**: Creates intelligent rules based on real exploit indicators
4. **Comprehensive Database Seeding**: Supports both bulk and incremental data updates

## Architecture

- **Backend**: FastAPI with SQLAlchemy ORM (`backend/main.py`)
- **Frontend**: React with Tailwind CSS (`frontend/src/App.js`)
- **Database**: PostgreSQL with enhanced schema:
  - `cves`: CVE information with PoC metadata and bulk processing fields
  - `sigma_rules`: Enhanced SIGMA rules with quality scoring and nomi-sec data
  - `rule_templates`: Template patterns for rule generation
  - `bulk_processing_jobs`: Job tracking for bulk operations
- **Data Processing**: 
  - `nvd_bulk_processor.py`: NVD JSON dataset downloader and processor
  - `nomi_sec_client.py`: nomi-sec PoC-in-GitHub API integration
  - `enhanced_sigma_generator.py`: Advanced SIGMA rule generation
  - `bulk_seeder.py`: Coordinated bulk seeding operations
- **Cache**: Redis (optional)
- **Deployment**: Docker Compose orchestration

## Common Development Commands

### Quick Start
```bash
# Recommended quick start
chmod +x start.sh
./start.sh

# Or using Make
make start
```

### Build and Run
```bash
# Build and start all services
docker-compose up -d --build

# Start individual services
docker-compose up -d db redis  # Database and cache only
docker-compose up -d backend   # Backend API
docker-compose up -d frontend  # React frontend
```

### Development Mode
```bash
# Using Make
make dev

# Or manually
docker-compose up -d db redis
cd backend && pip install -r requirements.txt && uvicorn main:app --reload
cd frontend && npm install && npm start
```

### Bulk Processing Commands
```bash
# Run bulk seeding standalone
cd backend && python bulk_seeder.py

# Bulk seed specific year range
cd backend && python -c "
import asyncio
from bulk_seeder import BulkSeeder
from main import SessionLocal
seeder = BulkSeeder(SessionLocal())
asyncio.run(seeder.full_bulk_seed(start_year=2020, end_year=2025))
"

# Incremental update only
cd backend && python -c "
import asyncio
from bulk_seeder import BulkSeeder
from main import SessionLocal
seeder = BulkSeeder(SessionLocal())
asyncio.run(seeder.incremental_update())
"
```

### Frontend Commands
```bash
cd frontend
npm install          # Install dependencies
npm start           # Development server (port 3000)
npm run build       # Production build
npm test            # Run tests
```

### Backend Commands
```bash
cd backend
pip install -r requirements.txt
uvicorn main:app --reload              # Development server (port 8000)
uvicorn main:app --host 0.0.0.0 --port 8000  # Production server
```

### Database Operations
```bash
# Connect to database
docker-compose exec db psql -U cve_user -d cve_sigma_db

# View logs
docker-compose logs -f backend
docker-compose logs -f frontend
```

### Other Make Commands
```bash
make stop           # Stop all services
make restart        # Restart all services
make logs          # View application logs
make clean         # Clean up containers and volumes
make setup         # Initial setup (creates .env from .env.example)
```

## Key Configuration

### Environment Variables (.env)
- `NVD_API_KEY`: Optional NVD API key for higher rate limits (5→50 requests/30s)
- `GITHUB_TOKEN`: Optional GitHub token for exploit analysis (enhances rule generation)
- `OPENAI_API_KEY`: Optional OpenAI API key for AI-enhanced SIGMA rule generation
- `ANTHROPIC_API_KEY`: Optional Anthropic API key for AI-enhanced SIGMA rule generation
- `OLLAMA_BASE_URL`: Optional Ollama base URL for local model AI-enhanced SIGMA rule generation
- `LLM_PROVIDER`: Optional LLM provider selection (openai, anthropic, ollama)
- `LLM_MODEL`: Optional LLM model selection (provider-specific)
- `DATABASE_URL`: PostgreSQL connection string
- `REACT_APP_API_URL`: Backend API URL for frontend

### Service URLs
- Frontend: http://localhost:3000
- Backend API: http://localhost:8000
- API Documentation: http://localhost:8000/docs
- Database: localhost:5432
- Redis: localhost:6379

### Enhanced API Endpoints

#### Bulk Processing
- `POST /api/bulk-seed` - Start complete bulk seeding (NVD + nomi-sec)
- `POST /api/incremental-update` - Update with NVD modified/recent feeds
- `POST /api/sync-nomi-sec` - Synchronize nomi-sec PoC data
- `POST /api/regenerate-rules` - Regenerate SIGMA rules with enhanced data
- `GET /api/bulk-jobs` - Get bulk processing job status
- `GET /api/bulk-status` - Get comprehensive system status
- `GET /api/poc-stats` - Get PoC-related statistics

#### Enhanced Data Access
- `GET /api/stats` - Enhanced statistics with PoC coverage
- `GET /api/claude-status` - Get Claude API availability status
- All existing CVE and SIGMA rule endpoints now include enhanced data fields

#### LLM-Enhanced Rule Generation
- `POST /api/llm-enhanced-rules` - Generate SIGMA rules using LLM AI analysis (supports multiple providers)
- `GET /api/llm-status` - Check LLM API availability and configuration for all providers
- `POST /api/llm-switch` - Switch between LLM providers and models

## Code Architecture Details

### Enhanced Backend Structure
- **main.py**: Core FastAPI application with enhanced endpoints
- **nvd_bulk_processor.py**: NVD JSON dataset downloader and processor
- **nomi_sec_client.py**: nomi-sec PoC-in-GitHub API integration
- **enhanced_sigma_generator.py**: Advanced SIGMA rule generation with PoC data
- **llm_client.py**: Multi-provider LLM integration using LangChain for AI-enhanced rule generation
- **bulk_seeder.py**: Coordinated bulk processing operations

### Database Models (Enhanced)
- **CVE**: Enhanced with `poc_count`, `poc_data`, `bulk_processed`, `data_source`
- **SigmaRule**: Enhanced with `poc_source`, `poc_quality_score`, `nomi_sec_data`
- **RuleTemplate**: Template patterns for rule generation
- **BulkProcessingJob**: Job tracking for bulk operations

### Frontend Structure (Enhanced)
- **Three Main Tabs**: Dashboard, CVEs, SIGMA Rules
- **Enhanced Dashboard**: PoC coverage statistics, data synchronization controls
- **Enhanced CVE/Rule Display**: PoC quality indicators, exploit-based tagging
- **Task Monitoring**: Via Flower dashboard (http://localhost:5555)

### Data Processing Flow
1. **Bulk Seeding**: NVD JSON downloads → Database storage → nomi-sec PoC sync → Enhanced rule generation
2. **Incremental Updates**: NVD modified feeds → Update existing data → Sync new PoCs
3. **Rule Enhancement**: PoC analysis → Indicator extraction → Template selection → Enhanced SIGMA rule
4. **LLM-Enhanced Generation**: PoC content analysis → Multi-provider LLM processing → Advanced SIGMA rule creation

## Development Notes

### Enhanced Rule Generation Logic
The application now uses an advanced rule generation process:
1. **CVE Analysis**: Extract metadata from NVD bulk data
2. **PoC Quality Assessment**: nomi-sec PoC analysis with star count, recency, quality tiers
3. **Advanced Indicator Extraction**: Processes, files, network, registry, commands from PoC repositories
4. **Template Selection**: Smart template matching based on PoC indicators and CVE characteristics
5. **Enhanced Rule Population**: Incorporate real exploit indicators with quality scoring
6. **MITRE ATT&CK Mapping**: Automatic technique identification based on indicators
7. **LLM AI Enhancement**: Optional multi-provider LLM integration for intelligent rule generation from PoC code analysis

### Quality Tiers
- **Excellent** (80+ points): High star count, recent updates, detailed descriptions
- **Good** (60-79 points): Moderate quality indicators
- **Fair** (40-59 points): Basic PoC with some quality indicators
- **Poor** (20-39 points): Minimal quality indicators
- **Very Poor** (<20 points): Low-quality PoCs

### Multi-Provider LLM Integration Features
- **Multiple LLM Providers**: Support for OpenAI, Anthropic, and Ollama (local models)
- **Dynamic Provider Switching**: Switch between providers and models through UI or API
- **Intelligent Code Analysis**: LLMs analyze actual exploit code from PoC repositories
- **Advanced Rule Generation**: Creates sophisticated SIGMA rules with proper syntax and logic
- **Contextual Understanding**: Interprets CVE descriptions and maps them to appropriate detection patterns
- **Automatic Validation**: Generated rules are validated for SIGMA syntax compliance
- **Fallback Mechanism**: Automatically falls back to template-based generation if LLM is unavailable
- **Enhanced Metadata**: Rules include generation method tracking for quality assessment
- **LangChain Integration**: Uses LangChain for robust LLM integration and prompt management

### Supported LLM Providers and Models

#### OpenAI
- **API Key**: Set `OPENAI_API_KEY` environment variable
- **Supported Models**: gpt-4o, gpt-4o-mini, gpt-4-turbo, gpt-3.5-turbo
- **Default Model**: gpt-4o-mini
- **Rate Limits**: Based on OpenAI API limits

#### Anthropic
- **API Key**: Set `ANTHROPIC_API_KEY` environment variable
- **Supported Models**: claude-3-5-sonnet-20241022, claude-3-haiku-20240307, claude-3-opus-20240229
- **Default Model**: claude-3-5-sonnet-20241022
- **Rate Limits**: Based on Anthropic API limits

#### Ollama (Local Models)
- **Setup**: Install Ollama locally and set `OLLAMA_BASE_URL` (default: http://localhost:11434)
- **Supported Models**: llama3.2, codellama, mistral, llama2 (any Ollama-compatible model)
- **Default Model**: llama3.2
- **Rate Limits**: No external API limits (local processing)

### Testing and Validation
- **Frontend tests**: `npm test` (in frontend directory)
- **Backend testing**: Use standalone scripts for bulk operations
- **API testing**: Use `/docs` endpoint for Swagger UI
- **Task Monitoring**: Monitor via Flower dashboard at http://localhost:5555
- **Celery Tasks**: Use `celery -A celery_config worker --loglevel=info` for debugging

### Security Considerations
- **API Keys**: Store NVD and GitHub tokens in environment variables
- **PoC Analysis**: Automated analysis of curated PoC repositories (safer than raw GitHub search)
- **Rate Limiting**: Built-in rate limiting for external APIs
- **Data Validation**: Enhanced validation for bulk data processing
- **Audit Trail**: Job tracking for all bulk operations

## Troubleshooting

### Common Issues
- **Bulk Processing Failures**: Check `/api/bulk-jobs` for detailed error messages
- **NVD Data Download Issues**: Verify NVD API key and network connectivity
- **nomi-sec API Timeouts**: Built-in retry logic, check network connectivity
- **Frontend build errors**: Run `npm install` in frontend directory
- **Database schema changes**: Restart backend to auto-create new tables
- **Memory issues during bulk processing**: Monitor system resources, consider smaller batch sizes

### Enhanced Rate Limits
- **NVD API**: 5 requests/30s (no key) → 50 requests/30s (with key)
- **nomi-sec API**: 1 request/second (built-in rate limiting)
- **GitHub API** (fallback): 60 requests/hour (no token) → 5000 requests/hour (with token)

### Performance Optimization
- **Bulk Processing**: Start with recent years (2020+) for faster initial setup
- **PoC Sync**: Use smaller batch sizes (50) for better stability
- **Rule Generation**: Monitor quality scores to prioritize high-value PoCs
- **Database**: Ensure proper indexing on CVE ID and PoC fields

### Monitoring
- **Frontend**: Use Bulk Jobs tab for real-time progress monitoring
- **Backend logs**: `docker-compose logs -f backend`
- **Job status**: Check `/api/bulk-status` for comprehensive system health
- **Database**: Monitor PoC coverage percentage and rule enhancement progress