import React, { useState, useEffect } from 'react';
import axios from 'axios';
import { Prism as SyntaxHighlighter } from 'react-syntax-highlighter';
import { tomorrow } from 'react-syntax-highlighter/dist/esm/styles/prism';
import './App.css';

const API_BASE_URL = process.env.REACT_APP_API_URL || 'http://localhost:8000';

function App() {
  const [cves, setCves] = useState([]);
  const [cveSearch, setCveSearch] = useState('');
  const [cveFilters, setCveFilters] = useState({ severity: '' });
  const [cvePagination, setCvePagination] = useState({ skip: 0, limit: 20, total: 0, hasMore: false });
  const [loadingCves, setLoadingCves] = useState(false);
  const [sigmaRules, setSigmaRules] = useState([]);
  const [selectedCve, setSelectedCve] = useState(null);
  const [stats, setStats] = useState({});
  const [loading, setLoading] = useState(true);
  const [activeTab, setActiveTab] = useState('dashboard');
  const [fetchingCves, setFetchingCves] = useState(false);
  const [testResult, setTestResult] = useState(null);
  const [pocStats, setPocStats] = useState({});
  const [gitHubPocStats, setGitHubPocStats] = useState({});
  const [exploitdbStats, setExploitdbStats] = useState({});
  const [cisaKevStats, setCisaKevStats] = useState({});
  const [llmStatus, setLlmStatus] = useState({});
  const [exploitSyncDropdownOpen, setExploitSyncDropdownOpen] = useState(false);

  // Function to fetch CVEs with search and pagination
  const fetchCves = async (search = '', filters = {}, pagination = { skip: 0, limit: 20 }) => {
    setLoadingCves(true);
    try {
      const params = new URLSearchParams({
        skip: pagination.skip.toString(),
        limit: pagination.limit.toString(),
      });
      
      if (search) params.append('search', search);
      if (filters.severity) params.append('severity', filters.severity);
      
      const response = await axios.get(`${API_BASE_URL}/api/cves?${params}`);
      setCves(response.data.cves || []);
      setCvePagination({
        skip: response.data.skip || 0,
        limit: response.data.limit || 20,
        total: response.data.total || 0,
        hasMore: response.data.has_more || false
      });
    } catch (error) {
      console.error('Error fetching CVEs:', error);
    } finally {
      setLoadingCves(false);
    }
  };

  useEffect(() => {
    fetchData();
  }, []);

  // Close dropdown when clicking outside
  useEffect(() => {
    const handleClickOutside = (event) => {
      if (exploitSyncDropdownOpen && !event.target.closest('.relative')) {
        setExploitSyncDropdownOpen(false);
      }
    };

    document.addEventListener('mousedown', handleClickOutside);
    return () => {
      document.removeEventListener('mousedown', handleClickOutside);
    };
  }, [exploitSyncDropdownOpen]);


  // Note: Scheduler functionality removed - now handled by Celery Beat
  // Monitoring available via Flower at http://localhost:5555

  const fetchData = async () => {
    try {
      setLoading(true);
      const [cvesRes, rulesRes, statsRes, pocStatsRes, githubPocStatsRes, exploitdbStatsRes, cisaKevStatsRes, llmStatusRes] = await Promise.all([
        axios.get(`${API_BASE_URL}/api/cves?limit=20`),
        axios.get(`${API_BASE_URL}/api/sigma-rules`),
        axios.get(`${API_BASE_URL}/api/stats`),
        axios.get(`${API_BASE_URL}/api/poc-stats`),
        axios.get(`${API_BASE_URL}/api/github-poc-stats`).catch(err => ({ data: {} })),
        axios.get(`${API_BASE_URL}/api/exploitdb-stats`).catch(err => ({ data: {} })),
        axios.get(`${API_BASE_URL}/api/cisa-kev-stats`).catch(err => ({ data: {} })),
        axios.get(`${API_BASE_URL}/api/llm-status`).catch(err => ({ data: {} }))
      ]);
      
      setCves(cvesRes.data.cves || []);
      setCvePagination({
        skip: cvesRes.data.skip || 0,
        limit: cvesRes.data.limit || 20,
        total: cvesRes.data.total || 0,
        hasMore: cvesRes.data.has_more || false
      });
      setSigmaRules(rulesRes.data);
      setStats(statsRes.data);
      setPocStats(pocStatsRes.data);
      setGitHubPocStats(githubPocStatsRes.data);
      setExploitdbStats(exploitdbStatsRes.data);
      setCisaKevStats(cisaKevStatsRes.data);
      setLlmStatus(llmStatusRes.data);
    } catch (error) {
      console.error('Error fetching data:', error);
    } finally {
      setLoading(false);
    }
  };


  const handleFetchCves = async () => {
    try {
      setFetchingCves(true);
      const response = await axios.post(`${API_BASE_URL}/api/fetch-cves`);
      console.log('Fetch response:', response.data);
      // Show success message and refresh after delay
      setTimeout(() => {
        fetchData();
        setFetchingCves(false);
      }, 5000); // Wait a bit longer for background task to complete
    } catch (error) {
      console.error('Error fetching CVEs:', error);
      setFetchingCves(false);
      // Show error state
      setTestResult({
        status: 'error',
        message: 'Failed to initiate CVE fetch. Check console logs.'
      });
    }
  };

  const testNvdConnection = async () => {
    try {
      const response = await axios.get(`${API_BASE_URL}/api/test-nvd`);
      setTestResult(response.data);
    } catch (error) {
      console.error('Error testing NVD connection:', error);
      setTestResult({
        status: 'error',
        message: 'Failed to test NVD connection'
      });
    }
  };

  const startBulkSeed = async (startYear = 2020, endYear = null) => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/bulk-seed`, {
        start_year: startYear,
        end_year: endYear,
        skip_nomi_sec: true
      });
      console.log('Bulk seed response:', response.data);
      // Refresh data immediately to show job started
      fetchData();
    } catch (error) {
      console.error('Error starting bulk seed:', error);
    }
  };

  const startIncrementalUpdate = async () => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/incremental-update`);
      console.log('Incremental update response:', response.data);
      fetchData();
    } catch (error) {
      console.error('Error starting incremental update:', error);
    }
  };

  const syncNomiSec = async (cveId = null) => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/sync-nomi-sec`, {
        cve_id: cveId
      });
      console.log('Nomi-sec sync response:', response.data);
      fetchData();
    } catch (error) {
      console.error('Error syncing nomi-sec:', error);
    }
  };

  const syncGitHubPocs = async (cveId = null) => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/sync-github-pocs`, {
        cve_id: cveId,
        batch_size: 50
      });
      console.log('GitHub PoC sync response:', response.data);
      
      // Show success message with Celery task info
      if (response.data.task_id) {
        console.log(`GitHub PoC sync task started: ${response.data.task_id}`);
        console.log(`Monitor at: ${response.data.monitor_url}`);
        
        // Show notification to user
        alert(`GitHub PoC sync started successfully!\nTask ID: ${response.data.task_id}\n\nMonitor progress at http://localhost:5555 (Flower Dashboard)`);
      }
      
      // Refresh data to show the task in bulk jobs
      fetchData();
    } catch (error) {
      console.error('Error syncing GitHub PoCs:', error);
      alert('Failed to start GitHub PoC sync. Please check the console for details.');
    }
  };

  const syncExploitDB = async (cveId = null) => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/sync-exploitdb`, {
        cve_id: cveId,
        batch_size: 30
      });
      console.log('ExploitDB sync response:', response.data);
      fetchData();
    } catch (error) {
      console.error('Error syncing ExploitDB:', error);
    }
  };

  const syncCISAKEV = async (cveId = null) => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/sync-cisa-kev`, {
        cve_id: cveId,
        batch_size: 100
      });
      console.log('CISA KEV sync response:', response.data);
      fetchData();
    } catch (error) {
      console.error('Error syncing CISA KEV:', error);
    }
  };

  const syncReferences = async () => {
    try {
      const response = await axios.post('http://localhost:8000/api/sync-references', {
        batch_size: 30,
        max_cves: 100,
        force_resync: false
      });
      console.log('Reference sync response:', response.data);
      fetchData();
    } catch (error) {
      console.error('Error syncing references:', error);
      alert('Error starting reference sync: ' + (error.response?.data?.detail || error.message));
    }
  };

  const regenerateRules = async (force = false) => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/regenerate-rules`, {
        force: force
      });
      console.log('Rule regeneration response:', response.data);
      fetchData();
    } catch (error) {
      console.error('Error regenerating rules:', error);
    }
  };

  const generateLlmRules = async (force = false) => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/llm-enhanced-rules`, {
        force: force
      });
      console.log('LLM rule generation response:', response.data);
      fetchData();
    } catch (error) {
      console.error('Error generating LLM-enhanced rules:', error);
    }
  };

  const switchLlmProvider = async (provider, model) => {
    try {
      const response = await axios.post(`${API_BASE_URL}/api/llm-switch`, {
        provider: provider,
        model: model
      });
      console.log('LLM provider switch response:', response.data);
      fetchData(); // Refresh to get updated status
    } catch (error) {
      console.error('Error switching LLM provider:', error);
      alert('Failed to switch LLM provider. Please check configuration.');
    }
  };

  const getSeverityColor = (severity) => {
    switch (severity?.toLowerCase()) {
      case 'critical': return 'bg-red-100 text-red-800';
      case 'high': return 'bg-orange-100 text-orange-800';
      case 'medium': return 'bg-yellow-100 text-yellow-800';
      case 'low': return 'bg-green-100 text-green-800';
      default: return 'bg-gray-100 text-gray-800';
    }
  };

  const formatDate = (dateString) => {
    return new Date(dateString).toLocaleDateString('en-US', {
      year: 'numeric',
      month: 'short',
      day: 'numeric'
    });
  };

  const Dashboard = () => (
    <div className="space-y-6">
      {/* Celery Monitoring Notice */}
      <div className="bg-blue-50 border border-blue-200 rounded-lg p-4">
        <div className="flex items-center">
          <div className="flex-shrink-0">
            <svg className="h-5 w-5 text-blue-400" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 20 20" fill="currentColor">
              <path fillRule="evenodd" d="M18 10a8 8 0 11-16 0 8 8 0 0116 0zm-7-4a1 1 0 11-2 0 1 1 0 012 0zM9 9a1 1 0 000 2v3a1 1 0 001 1h1a1 1 0 100-2v-3a1 1 0 00-1-1H9z" clipRule="evenodd" />
            </svg>
          </div>
          <div className="ml-3">
            <h3 className="text-sm font-medium text-blue-800">
              Task Scheduling & Monitoring
            </h3>
            <div className="mt-2 text-sm text-blue-700">
              <p>
                Automated tasks are now managed by Celery Beat. Monitor real-time task execution at{' '}
                <a 
                  href="http://localhost:5555" 
                  target="_blank" 
                  rel="noopener noreferrer"
                  className="font-medium underline hover:text-blue-900"
                >
                  Flower Dashboard
                </a>
              </p>
            </div>
          </div>
        </div>
      </div>
      
      <div className="grid grid-cols-1 md:grid-cols-5 gap-6">
        <div className="bg-white p-6 rounded-lg shadow">
          <h3 className="text-lg font-medium text-gray-900">Total CVEs</h3>
          <p className="text-3xl font-bold text-blue-600">{stats.total_cves || 0}</p>
          <p className="text-sm text-gray-500">Bulk: {stats.bulk_processed_cves || 0}</p>
        </div>
        <div className="bg-white p-6 rounded-lg shadow">
          <h3 className="text-lg font-medium text-gray-900">SIGMA Rules</h3>
          <p className="text-3xl font-bold text-green-600">{stats.total_sigma_rules || 0}</p>
          <p className="text-sm text-gray-500">Nomi-sec: {stats.nomi_sec_rules || 0}</p>
          <p className="text-sm text-gray-500">GitHub PoCs: {gitHubPocStats.github_poc_rules || 0}</p>
          <p className={`text-sm ${llmStatus.status === 'ready' ? 'text-green-600' : 'text-red-500'}`}>
            LLM: {llmStatus.current_provider?.provider || 'Not Available'}
          </p>
        </div>
        <div className="bg-white p-6 rounded-lg shadow">
          <h3 className="text-lg font-medium text-gray-900">CVEs with PoCs</h3>
          <p className="text-3xl font-bold text-purple-600">{stats.cves_with_pocs || 0}</p>
          <p className="text-sm text-gray-500">{(stats.poc_coverage || 0).toFixed(1)}% coverage</p>
          <p className="text-sm text-gray-500">GitHub PoCs: {gitHubPocStats.cves_with_github_pocs || 0}</p>
          <p className="text-sm text-gray-500">ExploitDB: {exploitdbStats.total_exploitdb_cves || 0}</p>
          <p className="text-sm text-gray-500">CISA KEV: {cisaKevStats.total_kev_cves || 0}</p>
        </div>
        <div className="bg-white p-6 rounded-lg shadow">
          <h3 className="text-lg font-medium text-gray-900">Recent CVEs (7d)</h3>
          <p className="text-3xl font-bold text-orange-600">{stats.recent_cves_7_days || 0}</p>
        </div>
        <div className="bg-white p-6 rounded-lg shadow">
          <h3 className="text-lg font-medium text-gray-900">High Quality PoCs</h3>
          <p className="text-3xl font-bold text-indigo-600">{pocStats.high_quality_cves || 0}</p>
          <p className="text-sm text-gray-500">Avg: {(pocStats.avg_poc_count || 0).toFixed(1)}</p>
          <p className="text-sm text-gray-500">GitHub: {(gitHubPocStats.average_quality_score || 0).toFixed(1)}</p>
          <p className="text-sm text-gray-500">ExploitDB: {exploitdbStats.total_exploits || 0} exploits</p>
          <p className="text-sm text-gray-500">CISA KEV: {(cisaKevStats.average_threat_score || 0).toFixed(1)} threat</p>
        </div>
      </div>

      {/* Data Synchronization Controls */}
      <div className="bg-white rounded-lg shadow p-6">
        <h2 className="text-xl font-bold text-gray-900 mb-6">Data Synchronization</h2>
        
        {/* Phase 1: CVE Data Syncing */}
        <div className="mb-6">
          <h3 className="text-lg font-medium text-gray-900 mb-3">Phase 1: CVE Data Syncing</h3>
          <div className="grid grid-cols-1 md:grid-cols-2 gap-4">
            <button
              onClick={() => startBulkSeed(2002)}
              className="px-4 py-2 rounded-md text-white bg-blue-600 hover:bg-blue-700"
            >
              Sync NVD CVEs
            </button>
            <button
              onClick={startIncrementalUpdate}
              className="px-4 py-2 rounded-md text-white bg-green-600 hover:bg-green-700"
            >
              Incremental Update
            </button>
          </div>
        </div>

        {/* Phase 2: Exploit Data Syncing */}
        <div className="mb-6">
          <h3 className="text-lg font-medium text-gray-900 mb-3">Phase 2: Exploit Data Syncing</h3>
          <div className="relative inline-block text-left">
            <button
              onClick={() => setExploitSyncDropdownOpen(!exploitSyncDropdownOpen)}
              className="inline-flex items-center justify-center w-full rounded-md border shadow-sm px-4 py-2 text-sm font-medium border-gray-300 bg-white text-gray-700 hover:bg-gray-50"
            >
              Sync Exploit Data
              <svg className="-mr-1 ml-2 h-5 w-5" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 20 20" fill="currentColor">
                <path fillRule="evenodd" d="M5.293 7.293a1 1 0 011.414 0L10 10.586l3.293-3.293a1 1 0 111.414 1.414l-4 4a1 1 0 01-1.414 0l-4-4a1 1 0 010-1.414z" clipRule="evenodd" />
              </svg>
            </button>

            {exploitSyncDropdownOpen && (
              <div className="origin-top-right absolute right-0 mt-2 w-56 rounded-md shadow-lg bg-white ring-1 ring-black ring-opacity-5 z-10">
                <div className="py-1">
                  <button
                    onClick={() => {
                      syncNomiSec();
                      setExploitSyncDropdownOpen(false);
                    }}
                    className="block w-full text-left px-4 py-2 text-sm text-gray-700 hover:bg-purple-50 hover:text-purple-900"
                  >
                    <span className="inline-block w-3 h-3 rounded-full mr-2 bg-purple-600"></span>
                    Sync nomi-sec PoCs
                  </button>
                  <button
                    onClick={() => {
                      syncGitHubPocs();
                      setExploitSyncDropdownOpen(false);
                    }}
                    className="block w-full text-left px-4 py-2 text-sm text-gray-700 hover:bg-green-50 hover:text-green-900"
                  >
                    <span className="inline-block w-3 h-3 rounded-full mr-2 bg-green-600"></span>
                    Sync GitHub PoCs
                  </button>
                  <button
                    onClick={() => {
                      syncExploitDB();
                      setExploitSyncDropdownOpen(false);
                    }}
                    className="block w-full text-left px-4 py-2 text-sm text-gray-700 hover:bg-red-50 hover:text-red-900"
                  >
                    <span className="inline-block w-3 h-3 rounded-full mr-2 bg-red-600"></span>
                    Sync ExploitDB
                  </button>
                  <button
                    onClick={() => {
                      syncCISAKEV();
                      setExploitSyncDropdownOpen(false);
                    }}
                    className="block w-full text-left px-4 py-2 text-sm text-gray-700 hover:bg-yellow-50 hover:text-yellow-900"
                  >
                    <span className="inline-block w-3 h-3 rounded-full mr-2 bg-yellow-600"></span>
                    Sync CISA KEV
                  </button>
                </div>
              </div>
            )}
          </div>
        </div>

        {/* Phase 3: Reference Data Syncing */}
        <div className="mb-6">
          <h3 className="text-lg font-medium text-gray-900 mb-3">Phase 3: Reference Data Syncing</h3>
          <button
            onClick={syncReferences}
            className="px-4 py-2 rounded-md text-white bg-orange-600 hover:bg-orange-700"
          >
            Sync References
          </button>
        </div>

        {/* Phase 4: Rule Generation */}
        <div>
          <h3 className="text-lg font-medium text-gray-900 mb-3">Phase 4: Rule Generation</h3>
          <div className="grid grid-cols-1 md:grid-cols-2 gap-4">
            <button
              onClick={() => regenerateRules()}
              className="px-4 py-2 rounded-md text-white bg-indigo-600 hover:bg-indigo-700"
            >
              Regenerate Rules
            </button>
            <button
              onClick={() => generateLlmRules()}
              disabled={llmStatus.status !== 'ready'}
              className={`px-4 py-2 rounded-md text-white ${
                llmStatus.status !== 'ready'
                  ? 'bg-gray-400 cursor-not-allowed' 
                  : 'bg-violet-600 hover:bg-violet-700'
              }`}
              title={llmStatus.status !== 'ready' ? 'LLM not configured' : ''}
            >
              Generate LLM Rules
            </button>
          </div>
        </div>
      </div>

      {/* LLM Configuration */}
      <div className="bg-white rounded-lg shadow p-6">
        <h2 className="text-xl font-bold text-gray-900 mb-4">LLM Configuration</h2>
        <div className="grid grid-cols-1 md:grid-cols-2 gap-6">
          <div>
            <h3 className="text-lg font-medium text-gray-900 mb-2">Current Provider</h3>
            <div className="space-y-2">
              <p className="text-sm text-gray-600">
                Provider: <span className="font-medium">{llmStatus.current_provider?.provider || 'Not configured'}</span>
              </p>
              <p className="text-sm text-gray-600">
                Model: <span className="font-medium">{llmStatus.current_provider?.model || 'Not configured'}</span>
              </p>
              <p className={`text-sm ${llmStatus.status === 'ready' ? 'text-green-600' : 'text-red-500'}`}>
                Status: <span className="font-medium">{llmStatus.status || 'Unknown'}</span>
              </p>
            </div>
          </div>
          <div>
            <h3 className="text-lg font-medium text-gray-900 mb-2">Available Providers</h3>
            <div className="space-y-2">
              {llmStatus.available_providers?.map(provider => (
                <div key={provider.name} className="flex items-center justify-between p-2 bg-gray-50 rounded">
                  <div>
                    <span className="font-medium">{provider.name}</span>
                    <span className={`ml-2 text-xs px-2 py-1 rounded ${
                      provider.available ? 'bg-green-100 text-green-800' : 'bg-red-100 text-red-800'
                    }`}>
                      {provider.available ? 'Available' : 'Not configured'}
                    </span>
                  </div>
                  {provider.available && provider.name !== llmStatus.current_provider?.provider && (
                    <button
                      onClick={() => switchLlmProvider(provider.name, provider.default_model)}
                      className="text-xs bg-blue-600 hover:bg-blue-700 text-white px-2 py-1 rounded"
                    >
                      Switch
                    </button>
                  )}
                </div>
              ))}
            </div>
          </div>
        </div>
      </div>

      <div className="bg-white rounded-lg shadow p-6">
        <div className="flex justify-between items-center mb-4">
          <h2 className="text-xl font-bold text-gray-900">Recent CVEs</h2>
          <div className="flex space-x-3">
            <button
              onClick={testNvdConnection}
              className="bg-green-600 hover:bg-green-700 text-white px-4 py-2 rounded-md text-sm"
            >
              Test NVD API
            </button>
            <button
              onClick={handleFetchCves}
              disabled={fetchingCves}
              className={`px-4 py-2 rounded-md text-white ${
                fetchingCves 
                  ? 'bg-gray-400 cursor-not-allowed' 
                  : 'bg-blue-600 hover:bg-blue-700'
              }`}
            >
              {fetchingCves ? 'Fetching...' : 'Fetch New CVEs'}
            </button>
          </div>
        </div>

        {testResult && (
          <div className={`mb-4 p-4 rounded-md ${
            testResult.status === 'success' ? 'bg-green-100 text-green-800' : 'bg-red-100 text-red-800'
          }`}>
            <div className="flex items-center">
              <span className="font-medium">NVD API Test: </span>
              <span className="ml-2">{testResult.message}</span>
            </div>
            {testResult.status === 'success' && (
              <div className="mt-2 text-sm">
                <p>✅ API Key: {testResult.has_api_key ? 'Present' : 'Not configured'}</p>
                <p>✅ Available results: {testResult.total_results || 0}</p>
              </div>
            )}
          </div>
        )}

        {fetchingCves && (
          <div className="mb-4 p-4 bg-blue-100 text-blue-800 rounded-md">
            <div className="flex items-center">
              <div className="animate-spin rounded-full h-4 w-4 border-b-2 border-blue-600 mr-2"></div>
              <span>Fetching CVEs from NVD API (30-day lookback)... This may take 1-2 minutes.</span>
            </div>
          </div>
        )}
        <div className="overflow-x-auto">
          <table className="min-w-full divide-y divide-gray-200">
            <thead className="bg-gray-50">
              <tr>
                <th className="px-6 py-3 text-left text-xs font-medium text-gray-500 uppercase tracking-wider">
                  CVE ID
                </th>
                <th className="px-6 py-3 text-left text-xs font-medium text-gray-500 uppercase tracking-wider">
                  Severity
                </th>
                <th className="px-6 py-3 text-left text-xs font-medium text-gray-500 uppercase tracking-wider">
                  CVSS Score
                </th>
                <th className="px-6 py-3 text-left text-xs font-medium text-gray-500 uppercase tracking-wider">
                  Published
                </th>
                <th className="px-6 py-3 text-left text-xs font-medium text-gray-500 uppercase tracking-wider">
                  Actions
                </th>
              </tr>
            </thead>
            <tbody className="bg-white divide-y divide-gray-200">
              {cves.slice(0, 10).map((cve) => (
                <tr key={cve.id}>
                  <td className="px-6 py-4 whitespace-nowrap text-sm font-medium text-gray-900">
                    {cve.cve_id}
                  </td>
                  <td className="px-6 py-4 whitespace-nowrap">
                    <span className={`inline-flex px-2 py-1 text-xs font-semibold rounded-full ${getSeverityColor(cve.severity)}`}>
                      {cve.severity || 'N/A'}
                    </span>
                  </td>
                  <td className="px-6 py-4 whitespace-nowrap text-sm text-gray-900">
                    {cve.cvss_score || 'N/A'}
                  </td>
                  <td className="px-6 py-4 whitespace-nowrap text-sm text-gray-500">
                    {cve.published_date ? formatDate(cve.published_date) : 'N/A'}
                  </td>
                  <td className="px-6 py-4 whitespace-nowrap text-sm font-medium">
                    <button
                      onClick={() => setSelectedCve(cve)}
                      className="text-blue-600 hover:text-blue-900"
                    >
                      View Details
                    </button>
                  </td>
                </tr>
              ))}
            </tbody>
          </table>
        </div>
      </div>
    </div>
  );

  const CVEList = () => {
    const handleSearch = (e) => {
      e.preventDefault();
      fetchCves(cveSearch, cveFilters, { skip: 0, limit: 20 });
    };

    const handleFilterChange = (filterName, value) => {
      const newFilters = { ...cveFilters, [filterName]: value };
      setCveFilters(newFilters);
      fetchCves(cveSearch, newFilters, { skip: 0, limit: 20 });
    };

    const handlePageChange = (newSkip) => {
      fetchCves(cveSearch, cveFilters, { skip: newSkip, limit: 20 });
    };

    return (
      <div className="bg-white rounded-lg shadow">
        <div className="px-6 py-4 border-b border-gray-200">
          <div className="flex justify-between items-center mb-4">
            <h2 className="text-xl font-bold text-gray-900">All CVEs</h2>
            <div className="text-sm text-gray-500">
              {cvePagination.total} total CVEs
            </div>
          </div>
          
          {/* Search and Filters */}
          <div className="flex gap-4 mb-4">
            <form onSubmit={handleSearch} className="flex-1">
              <div className="relative">
                <input
                  type="text"
                  value={cveSearch}
                  onChange={(e) => setCveSearch(e.target.value)}
                  placeholder="Search CVEs by ID, description, or affected products..."
                  className="w-full px-4 py-2 border border-gray-300 rounded-md focus:ring-2 focus:ring-blue-500 focus:border-transparent"
                />
                <button
                  type="submit"
                  className="absolute right-2 top-2 p-1 text-gray-400 hover:text-gray-600"
                >
                  <svg className="w-4 h-4" fill="none" stroke="currentColor" viewBox="0 0 24 24">
                    <path strokeLinecap="round" strokeLinejoin="round" strokeWidth={2} d="M21 21l-6-6m2-5a7 7 0 11-14 0 7 7 0 0114 0z" />
                  </svg>
                </button>
              </div>
            </form>
            
            <select
              value={cveFilters.severity}
              onChange={(e) => handleFilterChange('severity', e.target.value)}
              className="px-4 py-2 border border-gray-300 rounded-md focus:ring-2 focus:ring-blue-500 focus:border-transparent"
            >
              <option value="">All Severities</option>
              <option value="CRITICAL">Critical</option>
              <option value="HIGH">High</option>
              <option value="MEDIUM">Medium</option>
              <option value="LOW">Low</option>
            </select>
          </div>
        </div>
        
        {/* CVE List */}
        <div className="divide-y divide-gray-200">
          {loadingCves ? (
            <div className="p-6 text-center">
              <div className="animate-spin rounded-full h-8 w-8 border-b-2 border-blue-600 mx-auto"></div>
              <p className="mt-2 text-gray-500">Loading CVEs...</p>
            </div>
          ) : cves.length === 0 ? (
            <div className="p-6 text-center text-gray-500">
              No CVEs found matching your search criteria.
            </div>
          ) : (
            cves.map((cve) => (
              <div key={cve.id} className="p-6 hover:bg-gray-50">
                <div className="flex items-center justify-between">
                  <div className="flex-1">
                    <h3 className="text-lg font-medium text-gray-900">{cve.cve_id}</h3>
                    <p className="text-sm text-gray-600 mt-1 line-clamp-2">
                      {cve.description}
                    </p>
                    <div className="flex items-center mt-2 space-x-4">
                      <span className={`inline-flex px-2 py-1 text-xs font-semibold rounded-full ${getSeverityColor(cve.severity)}`}>
                        {cve.severity || 'N/A'}
                      </span>
                      <span className="text-sm text-gray-500">
                        CVSS: {cve.cvss_score || 'N/A'}
                      </span>
                      <span className="text-sm text-gray-500">
                        {cve.published_date ? formatDate(cve.published_date) : 'N/A'}
                      </span>
                      {cve.poc_count > 0 && (
                        <span className="inline-flex items-center px-2 py-1 text-xs font-semibold rounded-full bg-orange-100 text-orange-800">
                          🔍 {cve.poc_count} PoC{cve.poc_count > 1 ? 's' : ''}
                        </span>
                      )}
                    </div>
                  </div>
                  <button
                    onClick={() => setSelectedCve(cve)}
                    className="ml-4 bg-blue-600 hover:bg-blue-700 text-white px-4 py-2 rounded-md"
                  >
                    View Details
                  </button>
                </div>
              </div>
            ))
          )}
        </div>
        
        {/* Pagination */}
        {!loadingCves && cves.length > 0 && (
          <div className="px-6 py-4 border-t border-gray-200 flex items-center justify-between">
            <div className="text-sm text-gray-500">
              Showing {cvePagination.skip + 1} to {Math.min(cvePagination.skip + cvePagination.limit, cvePagination.total)} of {cvePagination.total} CVEs
            </div>
            <div className="flex space-x-2">
              <button
                onClick={() => handlePageChange(Math.max(0, cvePagination.skip - cvePagination.limit))}
                disabled={cvePagination.skip === 0}
                className="px-3 py-2 text-sm font-medium text-gray-500 bg-white border border-gray-300 rounded-md hover:bg-gray-50 disabled:opacity-50 disabled:cursor-not-allowed"
              >
                Previous
              </button>
              <button
                onClick={() => handlePageChange(cvePagination.skip + cvePagination.limit)}
                disabled={!cvePagination.hasMore}
                className="px-3 py-2 text-sm font-medium text-gray-500 bg-white border border-gray-300 rounded-md hover:bg-gray-50 disabled:opacity-50 disabled:cursor-not-allowed"
              >
                Next
              </button>
            </div>
          </div>
        )}
      </div>
    );
  };

  const SigmaRulesList = () => (
    <div className="bg-white rounded-lg shadow">
      <div className="px-6 py-4 border-b border-gray-200">
        <h2 className="text-xl font-bold text-gray-900">Generated SIGMA Rules</h2>
      </div>
      <div className="divide-y divide-gray-200">
        {sigmaRules.map((rule) => (
          <div key={rule.id} className="p-6">
            <div className="flex items-center justify-between mb-4">
              <div>
                <h3 className="text-lg font-medium text-gray-900">{rule.rule_name}</h3>
                <p className="text-sm text-gray-600">CVE: {rule.cve_id}</p>
                <div className="flex items-center mt-2 space-x-4">
                  <span className="inline-flex px-2 py-1 text-xs font-semibold rounded-full bg-blue-100 text-blue-800">
                    {rule.detection_type}
                  </span>
                  <span className="inline-flex px-2 py-1 text-xs font-semibold rounded-full bg-green-100 text-green-800">
                    {rule.confidence_level}
                  </span>
                  {rule.auto_generated && (
                    <span className="inline-flex px-2 py-1 text-xs font-semibold rounded-full bg-purple-100 text-purple-800">
                      Auto-generated
                    </span>
                  )}
                  {rule.exploit_based && (
                    <span className="inline-flex px-2 py-1 text-xs font-semibold rounded-full bg-red-100 text-red-800">
                      🔍 Exploit-Based
                    </span>
                  )}
                </div>
                {rule.github_repos && rule.github_repos.length > 0 && (
                  <div className="mt-2">
                    <p className="text-xs text-gray-500">
                      Based on {rule.github_repos.length} GitHub repository{rule.github_repos.length > 1 ? 's' : ''}:
                    </p>
                    <div className="flex flex-wrap gap-1 mt-1">
                      {rule.github_repos.slice(0, 3).map((repo, index) => (
                        <a
                          key={index}
                          href={repo}
                          target="_blank"
                          rel="noopener noreferrer"
                          className="text-xs text-blue-600 hover:text-blue-800 underline"
                        >
                          {repo.split('/').slice(-2).join('/')}
                        </a>
                      ))}
                      {rule.github_repos.length > 3 && (
                        <span className="text-xs text-gray-500">
                          +{rule.github_repos.length - 3} more
                        </span>
                      )}
                    </div>
                  </div>
                )}
              </div>
              <span className="text-sm text-gray-500">
                {formatDate(rule.created_at)}
              </span>
            </div>
            <div className="mt-4">
              <SyntaxHighlighter
                language="yaml"
                style={tomorrow}
                className="rounded-md"
                customStyle={{ fontSize: '12px' }}
              >
                {rule.rule_content}
              </SyntaxHighlighter>
            </div>
            {rule.exploit_indicators && (
              <div className="mt-4 p-3 bg-gray-50 rounded-md">
                <h4 className="text-sm font-medium text-gray-700 mb-2">Exploit Indicators Found:</h4>
                <ExploitIndicators indicators={rule.exploit_indicators} />
              </div>
            )}
          </div>
        ))}
      </div>
    </div>
  );

  const ExploitIndicators = ({ indicators }) => {
    try {
      const parsed = JSON.parse(indicators);
      return (
        <div className="space-y-2">
          {Object.entries(parsed).map(([category, items]) => (
            items.length > 0 && (
              <div key={category} className="flex flex-wrap items-center gap-2">
                <span className="text-xs font-medium text-gray-600 capitalize min-w-0">
                  {category}:
                </span>
                <div className="flex flex-wrap gap-1">
                  {items.slice(0, 5).map((item, index) => (
                    <span
                      key={index}
                      className="inline-flex px-2 py-1 text-xs rounded bg-gray-200 text-gray-700 font-mono"
                    >
                      {typeof item === 'string' && item.length > 30 ? item.substring(0, 30) + '...' : item}
                    </span>
                  ))}
                  {items.length > 5 && (
                    <span className="text-xs text-gray-500">+{items.length - 5} more</span>
                  )}
                </div>
              </div>
            )
          ))}
        </div>
      );
    } catch (e) {
      return <p className="text-xs text-gray-500">Invalid indicator data</p>;
    }
  };

  const CVEDetail = ({ cve, onClose }) => {
    const [cveRules, setCveRules] = useState([]);

    useEffect(() => {
      if (cve) {
        fetchCveRules(cve.cve_id);
      }
    }, [cve]);

    const fetchCveRules = async (cveId) => {
      try {
        const response = await axios.get(`${API_BASE_URL}/api/sigma-rules/${cveId}`);
        setCveRules(response.data);
      } catch (error) {
        console.error('Error fetching CVE rules:', error);
      }
    };

    return (
      <div className="fixed inset-0 bg-gray-600 bg-opacity-50 overflow-y-auto h-full w-full z-50">
        <div className="relative top-20 mx-auto p-5 border w-11/12 max-w-4xl shadow-lg rounded-md bg-white">
          <div className="flex justify-between items-center mb-4">
            <h2 className="text-2xl font-bold text-gray-900">{cve.cve_id}</h2>
            <button
              onClick={onClose}
              className="text-gray-400 hover:text-gray-600"
            >
              <svg className="w-6 h-6" fill="none" stroke="currentColor" viewBox="0 0 24 24">
                <path strokeLinecap="round" strokeLinejoin="round" strokeWidth={2} d="M6 18L18 6M6 6l12 12" />
              </svg>
            </button>
          </div>

          <div className="space-y-6">
            <div>
              <h3 className="text-lg font-medium text-gray-900 mb-2">Details</h3>
              <div className="bg-gray-50 p-4 rounded-md">
                <div className="grid grid-cols-2 gap-4">
                  <div>
                    <span className="text-sm font-medium text-gray-500">Severity:</span>
                    <span className={`ml-2 inline-flex px-2 py-1 text-xs font-semibold rounded-full ${getSeverityColor(cve.severity)}`}>
                      {cve.severity || 'N/A'}
                    </span>
                  </div>
                  <div>
                    <span className="text-sm font-medium text-gray-500">CVSS Score:</span>
                    <span className="ml-2 text-sm text-gray-900">{cve.cvss_score || 'N/A'}</span>
                  </div>
                  <div>
                    <span className="text-sm font-medium text-gray-500">Published:</span>
                    <span className="ml-2 text-sm text-gray-900">
                      {cve.published_date ? formatDate(cve.published_date) : 'N/A'}
                    </span>
                  </div>
                </div>
              </div>
            </div>

            <div>
              <h3 className="text-lg font-medium text-gray-900 mb-2">Description</h3>
              <p className="text-sm text-gray-700 bg-gray-50 p-4 rounded-md">
                {cve.description}
              </p>
            </div>

            {cve.affected_products && cve.affected_products.length > 0 && (
              <div>
                <h3 className="text-lg font-medium text-gray-900 mb-2">Affected Products</h3>
                <div className="bg-gray-50 p-4 rounded-md">
                  <ul className="list-disc list-inside space-y-1">
                    {cve.affected_products.slice(0, 5).map((product, index) => (
                      <li key={index} className="text-sm text-gray-700">{product}</li>
                    ))}
                    {cve.affected_products.length > 5 && (
                      <li className="text-sm text-gray-500">... and {cve.affected_products.length - 5} more</li>
                    )}
                  </ul>
                </div>
              </div>
            )}

            {/* References Section */}
            {cve.reference_urls && cve.reference_urls.length > 0 && (
              <div>
                <h3 className="text-lg font-medium text-gray-900 mb-2">References</h3>
                <div className="bg-gray-50 p-4 rounded-md">
                  <ul className="space-y-2">
                    {cve.reference_urls.map((url, index) => (
                      <li key={index}>
                        <a
                          href={url}
                          target="_blank"
                          rel="noopener noreferrer"
                          className="text-sm text-blue-600 hover:text-blue-800 underline break-all"
                        >
                          {url}
                        </a>
                      </li>
                    ))}
                  </ul>
                </div>
              </div>
            )}

            {/* Exploit/PoC Links Section */}
            {cve.poc_data && Object.keys(cve.poc_data).length > 0 && (
              <div>
                <h3 className="text-lg font-medium text-gray-900 mb-2">Exploits & Proof of Concepts</h3>
                <div className="space-y-4">
                  {Object.entries(cve.poc_data).map(([source, data]) => (
                    <div key={source} className="border border-gray-200 rounded-md p-4">
                      <div className="flex items-center justify-between mb-3">
                        <h4 className="font-medium text-gray-900 capitalize">{source}</h4>
                        <span className="inline-flex items-center px-2 py-1 text-xs font-semibold rounded-full bg-blue-100 text-blue-800">
                          {data.exploits?.length || data.pocs?.length || 0} items
                        </span>
                      </div>
                      
                      {(data.exploits || data.pocs || []).slice(0, 5).map((item, index) => (
                        <div key={index} className="mb-3 p-3 bg-gray-50 rounded">
                          <div className="flex items-start justify-between">
                            <div className="flex-1">
                              <h5 className="text-sm font-medium text-gray-900">
                                {item.title || item.name || item.description || 'Untitled'}
                              </h5>
                              {item.description && (
                                <p className="text-xs text-gray-600 mt-1 line-clamp-2">
                                  {item.description}
                                </p>
                              )}
                              <div className="flex items-center mt-2 space-x-3">
                                {item.html_url && (
                                  <a
                                    href={item.html_url}
                                    target="_blank"
                                    rel="noopener noreferrer"
                                    className="text-xs text-blue-600 hover:text-blue-800 underline"
                                  >
                                    View Source
                                  </a>
                                )}
                                {item.quality_analysis && (
                                  <span className={`inline-flex px-2 py-1 text-xs font-semibold rounded-full ${
                                    item.quality_analysis.quality_tier === 'excellent' ? 'bg-green-100 text-green-800' :
                                    item.quality_analysis.quality_tier === 'good' ? 'bg-blue-100 text-blue-800' :
                                    item.quality_analysis.quality_tier === 'fair' ? 'bg-yellow-100 text-yellow-800' :
                                    'bg-gray-100 text-gray-800'
                                  }`}>
                                    {item.quality_analysis.quality_tier}
                                  </span>
                                )}
                                {item.stargazers_count && (
                                  <span className="text-xs text-gray-500">
                                    ⭐ {item.stargazers_count}
                                  </span>
                                )}
                              </div>
                            </div>
                          </div>
                        </div>
                      ))}
                      
                      {(data.exploits || data.pocs || []).length > 5 && (
                        <p className="text-xs text-gray-500 mt-2">
                          ... and {(data.exploits || data.pocs || []).length - 5} more items
                        </p>
                      )}
                    </div>
                  ))}
                </div>
              </div>
            )}

            <div>
              <h3 className="text-lg font-medium text-gray-900 mb-2">Generated SIGMA Rules ({cveRules.length})</h3>
              {cveRules.length > 0 ? (
                <div className="space-y-4">
                  {cveRules.map((rule) => (
                    <div key={rule.id} className="border border-gray-200 rounded-md p-4">
                      <div className="flex justify-between items-center mb-2">
                        <h4 className="font-medium text-gray-900">{rule.rule_name}</h4>
                        <div className="flex space-x-2">
                          <span className="inline-flex px-2 py-1 text-xs font-semibold rounded-full bg-blue-100 text-blue-800">
                            {rule.detection_type}
                          </span>
                          <span className="inline-flex px-2 py-1 text-xs font-semibold rounded-full bg-green-100 text-green-800">
                            {rule.confidence_level}
                          </span>
                          {rule.exploit_based && (
                            <span className="inline-flex px-2 py-1 text-xs font-semibold rounded-full bg-red-100 text-red-800">
                              🔍 Exploit-Based
                            </span>
                          )}
                        </div>
                      </div>
                      {rule.github_repos && rule.github_repos.length > 0 && (
                        <div className="mb-3 p-2 bg-blue-50 rounded">
                          <p className="text-xs text-blue-700 font-medium mb-1">
                            Based on GitHub exploit analysis:
                          </p>
                          <div className="flex flex-wrap gap-1">
                            {rule.github_repos.slice(0, 2).map((repo, index) => (
                              <a
                                key={index}
                                href={repo}
                                target="_blank"
                                rel="noopener noreferrer"
                                className="text-xs text-blue-600 hover:text-blue-800 underline"
                              >
                                {repo.split('/').slice(-2).join('/')}
                              </a>
                            ))}
                            {rule.github_repos.length > 2 && (
                              <span className="text-xs text-blue-600">
                                +{rule.github_repos.length - 2} more
                              </span>
                            )}
                          </div>
                        </div>
                      )}
                      <SyntaxHighlighter
                        language="yaml"
                        style={tomorrow}
                        className="rounded-md"
                        customStyle={{ fontSize: '12px' }}
                      >
                        {rule.rule_content}
                      </SyntaxHighlighter>
                      {rule.exploit_indicators && (
                        <div className="mt-3 p-2 bg-gray-50 rounded">
                          <p className="text-xs font-medium text-gray-700 mb-1">Exploit Indicators:</p>
                          <ExploitIndicators indicators={rule.exploit_indicators} />
                        </div>
                      )}
                    </div>
                  ))}
                </div>
              ) : (
                <p className="text-gray-500 bg-gray-50 p-4 rounded-md">No SIGMA rules generated for this CVE yet.</p>
              )}
            </div>
          </div>
        </div>
      </div>
    );
  };


  // Note: SchedulerManager component removed - job scheduling now handled by Celery Beat
  // Task monitoring available via Flower dashboard at http://localhost:5555

  if (loading) {
    return (
      <div className="min-h-screen bg-gray-100 flex items-center justify-center">
        <div className="text-center">
          <div className="animate-spin rounded-full h-32 w-32 border-b-2 border-blue-600"></div>
          <p className="mt-4 text-gray-600">Loading...</p>
        </div>
      </div>
    );
  }

  return (
    <div className="min-h-screen bg-gray-100">
      <nav className="bg-white shadow">
        <div className="max-w-7xl mx-auto px-4 sm:px-6 lg:px-8">
          <div className="flex justify-between h-16">
            <div className="flex">
              <div className="flex-shrink-0 flex items-center">
                <h1 className="text-xl font-bold text-gray-900">CVE-SIGMA Auto Generator</h1>
              </div>
              <div className="ml-6 flex space-x-8">
                <button
                  onClick={() => setActiveTab('dashboard')}
                  className={`inline-flex items-center px-1 pt-1 border-b-2 text-sm font-medium ${
                    activeTab === 'dashboard'
                      ? 'border-blue-500 text-gray-900'
                      : 'border-transparent text-gray-500 hover:text-gray-700 hover:border-gray-300'
                  }`}
                >
                  Dashboard
                </button>
                <button
                  onClick={() => setActiveTab('cves')}
                  className={`inline-flex items-center px-1 pt-1 border-b-2 text-sm font-medium ${
                    activeTab === 'cves'
                      ? 'border-blue-500 text-gray-900'
                      : 'border-transparent text-gray-500 hover:text-gray-700 hover:border-gray-300'
                  }`}
                >
                  CVEs
                </button>
                <button
                  onClick={() => setActiveTab('rules')}
                  className={`inline-flex items-center px-1 pt-1 border-b-2 text-sm font-medium ${
                    activeTab === 'rules'
                      ? 'border-blue-500 text-gray-900'
                      : 'border-transparent text-gray-500 hover:text-gray-700 hover:border-gray-300'
                  }`}
                >
                  SIGMA Rules
                </button>
              </div>
            </div>
          </div>
        </div>
      </nav>

      <main className="max-w-7xl mx-auto py-6 sm:px-6 lg:px-8">
        <div className="px-4 py-6 sm:px-0">
          {activeTab === 'dashboard' && <Dashboard />}
          {activeTab === 'cves' && <CVEList />}
          {activeTab === 'rules' && <SigmaRulesList />}
        </div>
      </main>

      {selectedCve && (
        <CVEDetail cve={selectedCve} onClose={() => setSelectedCve(null)} />
      )}
    </div>
  );
}

export default App;
